#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdbool.h>

#include <signal.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <sys/epoll.h>
#include <curl/curl.h>

#include "curl/multi.h"
#include "http.h"
#include "util.h"
#include "crawler.h"
#include "json.h"
#include "module.h"

#define MAX_FAILURE_CNT 5
#define MAX_CONNECTIONS 32
#define MAX_CONNECTIONS_PER_HOST 5
#define EPOLL_BUF_SIZE MAX_CONNECTIONS
#define DEATH_TIMEOUT_S 30

volatile sig_atomic_t shoulddie = 0;
volatile struct timespec fatalsigrecved = { 0 };

void fatalsighandler(int status) {
    struct timespec now;
    // clock_gettime is safe to call within a signal handler
    if (clock_gettime(CLOCK_MONOTONIC, &now) < 0)
        // die calls fprintf which is not safe to call within a signal handler, but we're going
        // to die anyways so does it really matter?
        die("clock_gettime failed: %s (code %d)", strerror(errno), errno);
    if (shoulddie == 0) {
        fatalsigrecved = now;
        shoulddie = 1;
    }
    if (now.tv_sec - fatalsigrecved.tv_sec > DEATH_TIMEOUT_S)
        die("death timeout exceeded (%ds)", DEATH_TIMEOUT_S);
}

int perf_epoll_wait (int __epfd, struct epoll_event *__events,
		       int __maxevents, int __timeout) {
    return epoll_wait(__epfd, __events, __maxevents, __timeout);
}

CURLMcode perf_curl_multi_socket_action(CURLM *multi_handle,
                                               curl_socket_t s,
                                               int ev_bitmask,
                                               int *running_handles) {
    return curl_multi_socket_action(multi_handle, s, ev_bitmask, running_handles);
}

bool is_allowed_host(const char *host, const char **allowed_hosts) {
    if (allowed_hosts == NULL)
        return true;
    for (int i = 0; allowed_hosts[i] != NULL; i++)
        if (strcmp(host, allowed_hosts[i]) == 0)
            return true;
    return false;
}

CURLUcode append2tocrawl(CURLU *curl_url_h, char *link, const char **allowed_hosts, charp_deque_t *to_crawl) {
    CURLUcode url_res;
    url_res = curl_url_set(curl_url_h, CURLUPART_URL, link, 0);
    if (url_res != CURLUE_OK){
        char sanitized[100];
        sanitize2ascii(sanitized, link, sizeof(sanitized));
        error("URL parsing failed for \"%s\": %s (code %d)", 
                sanitized, curl_url_strerror(url_res), url_res);
        return url_res;
    } 
    char *curl_link_host = NULL;
    url_res = curl_url_get(curl_url_h, CURLUPART_HOST, &curl_link_host, CURLU_PUNYCODE);
    if (url_res != CURLUE_OK) {
        error("URL host parsing failed: %s (code %d)", curl_url_strerror(url_res), url_res);
        return url_res;
    }
    if (!is_allowed_host(curl_link_host, allowed_hosts)) {
        char sanitized[100];
        sanitize2ascii(sanitized, link, sizeof(sanitized));
        error("URL not in allowed hosts \"%s\", not crawling", sanitized);
        curl_free(curl_link_host);
        return CURLUE_BAD_HOSTNAME;
    }
    curl_free(curl_link_host);
    deque_push_back(*to_crawl, link);
    return CURLUE_OK;
}

#define MAX_SOCKETS_PER_TRANS 4

typedef struct transfer {
    struct transfer *next;
    CURL *handle;
    cbdata_t cbdata;
    char *url;
    int hostentryind;
} transfer_t;

typedef struct {
    const crawlerconfig_t *config;
    charp_deque_t to_crawl;
    charp_dynarr_t visited;
    host_dynarr_t hostentries;
    transfer_t *headtransfer;
    int epollfd;
    long timeout_ms;
    CURLM *curl_multi_h;
    int_dynarr_t sockets;
    moduleentryp_dynarr_t modules;
} crawlerstate_t;

typedef enum {
    TRANS_OK,
    TRANS_ERROR,
    TRANS_VISITED,
    TRANS_HOST_ERROR_EXCEEDED,
    TRANS_HOST_ERROR,
    TRANS_WRITE_ERR,
    TRANS_FATAL,
} transfercode_t;

int socketcb(CURL *easy, curl_socket_t s, int what, crawlerstate_t *clientp, transfer_t *socketp) {
    //printf("socketcb(%p, %d, %d, %p, %p)\n", easy, s, what, (void*)clientp, (void*)socketp);
    struct epoll_event event = { .data = { .fd = s } };
    int sockind = -1;
    for (size_t i = 0; i < clientp->sockets.len; i++) {
        if (s == clientp->sockets.data[i]) {
            sockind = i;
            break;
        }
    }
    switch (what) {
    case CURL_POLL_NONE:
        return 0;
    case CURL_POLL_IN:
        event.events = EPOLLIN;
        break;
    case CURL_POLL_OUT:
        event.events = EPOLLOUT;
        break;
    case CURL_POLL_INOUT:
        event.events = EPOLLIN | EPOLLOUT;
        break;
    case CURL_POLL_REMOVE:
        if (sockind < 0) {
            fatal("invalid socket specified!");
            return -1;
        }
        if (epoll_ctl(clientp->epollfd, EPOLL_CTL_DEL, s, NULL) < 0) {
            fatal("epoll_ctl failed: %s (code %d)", strerror(errno), errno);
            return -1;
        }
        dynarr_remove(clientp->sockets, sockind);
        return 0;
    }
    int op = sockind < 0 ? EPOLL_CTL_ADD : EPOLL_CTL_MOD;
    if (epoll_ctl(clientp->epollfd, op, s, &event) < 0) {
        fatal("epoll_ctl failed: %s (code %d)", strerror(errno), errno);
        return -1;
    }
    if (op == EPOLL_CTL_ADD)
        dynarr_push(clientp->sockets, s);
    return 0;
}

int timercb(CURLM *multi, long timeout_ms, long *clientp) {
    *clientp = timeout_ms;
    return 0;
}

bool gethostdata(const char *host, const char **avail_uas, host_dynarr_t *host_data, int *hostentryind) {
    const char *picked_ua = NULL;
    for (size_t i = 0; i < host_data->len; i++) {
        if (strcmp(host, host_data->data[i].host) == 0) {
            *hostentryind = i;
            return true;
        }
    }
    if (avail_uas == NULL) {
        *hostentryind = -1;
        return false;
    }
    // pick one
    size_t navail_uas = 0;
    for (; avail_uas[navail_uas] != NULL; navail_uas++)
        ;
    picked_ua = avail_uas[(size_t) random() % navail_uas];
    const char prefix[] = "User-Agent: ";
    char *new_buf = xmalloc(strlen(picked_ua) + sizeof(prefix));
    *new_buf = '\0';
    strcpy(new_buf, prefix);
    strcat(new_buf, picked_ua);
    struct curl_slist *headers = NULL;
    if ((headers = curl_slist_append(headers, new_buf)) == NULL) {
        xfree(new_buf);
        curl_slist_free_all(headers);
        *hostentryind = -1;
        return false;
    }
    hostentry_t new_entry = { 
        .host = host, 
        .headers = headers, 
        .totalfailurecnt = 0, 
        .failurecnt = 0, 
        .visitcnt = 0 
    };
    dynarr_push(*host_data, new_entry);
    *hostentryind = host_data->len - 1;
    return false;
}

bool didvisit(const char *url, charp_dynarr_t *visited) {
    int lhslen = strlen(url);
    for (size_t i = 0; i < visited->len; i++) {
        int rhslen = strlen(visited->data[i]);
        if (lhslen != rhslen)
            continue;
        if (memcmp(url, visited->data[i], lhslen) == 0)
            return true;
    }
    return false;
}

transfercode_t starttransfer(CURLU *curl_url_h, crawlerstate_t *state, char *url, 
        transfer_t **transfer_ret, bool wasrequested) {
    char sanitized[100];
    if (didvisit(url, &state->visited))
        // DO NOT free the URL. We are going to use it in our JSON later
        return TRANS_VISITED;
    dynarr_push(state->visited, url);
    sanitize2ascii(sanitized, url, sizeof(sanitized));
    info("crawling \"%s\"", sanitized);

    // Retrieves handle and host data
    *transfer_ret = xmalloc(sizeof(transfer_t));
    memset(*transfer_ret, 0, sizeof(transfer_t));
    transfer_t *transfer = *transfer_ret;
    char *curl_host = NULL;
    CURLUcode url_res;
    if ((url_res = curl_url_set(curl_url_h, CURLUPART_URL, url, 0)) != CURLUE_OK ||
        (url_res = curl_url_get(curl_url_h, CURLUPART_HOST, &curl_host, CURLU_PUNYCODE) != CURLUE_OK)) {
        fatal("URL host parsing failed: %s (code %d)", curl_url_strerror(url_res), url_res);
        curl_free(curl_host);
        xfree(transfer);
        return TRANS_FATAL;
    }
    int hostentryind;
    bool wascached = gethostdata(curl_host, useragents, &state->hostentries, &hostentryind);
    if (hostentryind < 0) {
        // host is (should be) in punycode so it's fine if we don't sanitize it
        fatal("Failed to get host entry for host \"%s\"", curl_host); 
        curl_free(curl_host);
        xfree(transfer);
        return TRANS_FATAL;
    }
    if (wascached)
        curl_free(curl_host);
    hostentry_t *hostentry = &state->hostentries.data[hostentryind];
    initcbdata(url, &state->modules, &transfer->cbdata);
    transfer->hostentryind = hostentryind;
    transfer->url = url;
    transfer->handle = makehandle(url, hostentry, &transfer->cbdata, wasrequested);
    if (transfer->handle == NULL) {
        fatal("makehandle() failed");
        xfree(transfer->cbdata.writecb_data.base);
        xfree(transfer);
        return TRANS_FATAL;
    }
    if (hostentry->failurecnt > MAX_FAILURE_CNT) {
        error("Max failure count (%d) for host exceeded", MAX_FAILURE_CNT); 
        curl_easy_cleanup(transfer->handle);
        xfree(transfer->cbdata.writecb_data.base);
        xfree(transfer);
        return TRANS_HOST_ERROR_EXCEEDED;
    }

    // Adds the transfer to be the multi handle tbh idk what im doing
    CURLMcode mc;
    mc = curl_multi_add_handle(state->curl_multi_h, transfer->handle);
    if (mc) {
        // erm,what the sigma?
        fatal("curl_multi_add_handle failed: %s (code %d)", curl_multi_strerror(mc), mc);
        curl_easy_cleanup(transfer->handle);
        xfree(transfer->cbdata.writecb_data.base);
        xfree(transfer);
        return TRANS_FATAL;
    }

    // Add the transfer to the end of the transfer list
    if (state->headtransfer == NULL) {
        state->headtransfer = transfer;
    } 
    else {
        transfer_t *last = state->headtransfer;
        for (; last->next != NULL; last = last->next)
            ;
        last->next = transfer;
    }

    return TRANS_OK;
}

void destroytransfer(CURLM *curl_multi_h, transfer_t *trans) {
    curl_multi_remove_handle(curl_multi_h, trans->handle);
    curl_easy_cleanup(trans->handle);
    if (trans->cbdata.writecb_data.base != NULL)
        xfree(trans->cbdata.writecb_data.base);
    xfree(trans);
}

// TODO: Extract arguments into a struct
transfercode_t transferfinished(CURLU *curl_url_h, crawlerstate_t *state, 
        CURLMsg *msg, transfer_t *trans, charp_dynarr_t *links, jsonkv_dynarr_t *extrajson,
        reqcb_t cb, void *cb_userdata) {
    CURLcode trans_res = msg->data.result;
    char sanitized[100];
    hostentry_t *hostentry = &state->hostentries.data[trans->hostentryind];
    hostentry->visitcnt++;
    // Check if transfer went OK
    if (trans_res != CURLE_OK) {
        sanitize2ascii(sanitized, trans->url, sizeof(sanitized));
        transfercode_t ret;
        headercb_data_t *headerdata;
        switch (trans_res) {
        case CURLE_WRITE_ERROR:
            headerdata = &trans->cbdata.headercb_data;
            if (!(headerdata->flags & HEADERCB_VALID_MIME))
                info("request failed to \"%s\": is not of mime type text/html, not crawling", sanitized);
            else if (headerdata->status != 200 && headerdata->status > 0)
                info("request failed to \"%s\": returned status code %d", headerdata->status, sanitized);
            else
                info("request failed to \"%s\": header parsing error or page too big", sanitized);
            ret = TRANS_WRITE_ERR;
            break;
        case CURLE_REMOTE_ACCESS_DENIED:
        case CURLE_BAD_CONTENT_ENCODING:
        case CURLE_PEER_FAILED_VERIFICATION:
        case CURLE_WEIRD_SERVER_REPLY:
        case CURLE_BAD_DOWNLOAD_RESUME:
            hostentry->failurecnt++;
            hostentry->totalfailurecnt++;
        case CURLE_RANGE_ERROR:
        case CURLE_UNSUPPORTED_PROTOCOL:
        case CURLE_AUTH_ERROR:
        case CURLE_LOGIN_DENIED:
        case CURLE_TOO_MANY_REDIRECTS:
        case CURLE_FILESIZE_EXCEEDED:
        case CURLE_HTTP2:
        case CURLE_HTTP3:
        case CURLE_HTTP2_STREAM:
        case CURLE_QUIC_CONNECT_ERROR:
            error("non-fatal error: %s (code %d)", curl_easy_strerror(trans_res), trans_res);
            ret = TRANS_ERROR;
            break;
        case CURLE_SSL_CONNECT_ERROR:
        case CURLE_COULDNT_RESOLVE_HOST:
        case CURLE_COULDNT_CONNECT:
        case CURLE_OPERATION_TIMEDOUT:
            info("retrying (eventually)...: %s (code %d)", curl_easy_strerror(trans_res), trans_res);
            // Add to to_crawl and remove from visited so that we retry the transfer (eventually)
            deque_push_front(state->to_crawl, trans->url);
            for (size_t i = 0; i < state->visited.len; i++) {
                // Shallow compare is fine
                if (state->visited.data[i] == trans->url) {
                    dynarr_remove(state->visited, i);
                    break;
                }
            }
            hostentry->failurecnt++;
            hostentry->totalfailurecnt++;
            ret = TRANS_HOST_ERROR;
            break;
        default:
            fatal("aborting...: %s (code %d)", curl_easy_strerror(trans_res), trans_res);
            ret = TRANS_FATAL;
            break;
        }
        return ret;
    }
    hostentry->failurecnt = 0;

    // Link aggregation and parsing logic
    writecb_data_t writecb_data = trans->cbdata.writecb_data; 
    char *page = (char*)writecb_data.base;
    size_t npage = writecb_data.begin - writecb_data.base;
    dynarr_init(charp_dynarr_t, *links);
    debug("%ld links in %ld bytes ", links->len, npage);
    pagecompletedata_t moduledata = { 
            .url = trans->url, 
            .handle = trans->handle, 
            .page = page, 
            .npage = npage, 
            .parsedlinks = links, 
    };

    // Call callback and terminate, if provided 
    if (cb != NULL) {
        dynarr_init(jsonkv_dynarr_t, *extrajson);
        cb(cb_userdata, moduledata.url, moduledata.page, moduledata.npage, trans->handle);
        return TRANS_OK;
    }

    // Module and JSON stuff
    dynarr_init(extradata_dynarr_t, moduledata.extradata);
    int rc;
    for (size_t i = 0; i < state->modules.len; i++) {
        if (state->modules.data[i]->module.onpagecomplete != NULL) {
            rc = state->modules.data[i]->module.onpagecomplete(
                state->modules.data[i]->module.userdata, &moduledata);
            if (rc != 0)
                error("module %s onpagecomplete failed with code %d", state->modules.data[i]->name, rc);
        }
    }
    for (size_t i = 0; i < links->len; i++)
        append2tocrawl(curl_url_h, links->data[i], state->config->allowedhosts, &state->to_crawl);
    for (size_t i = state->modules.len - 1; i != SIZE_MAX; i--) {
        if (state->modules.data[i]->module.onpagedestroy != NULL) {
            int rc;
            rc = state->modules.data[i]->module.onpagedestroy(
                state->modules.data[i]->module.userdata, &moduledata);
            if (rc != 0)
                error("module %s onpagedestroy failed with code %d", state->modules.data[i]->name, rc);
        }
    }
    jsonkv_dynarr_t *jsoncand = 
        searchextradata(EXTRA_JSON, "json", moduledata.extradata.data, moduledata.extradata.len);
    dynarr_destroy(moduledata.extradata);
    // TODO: Make it so we don't initialize an empty array on no extrajson
    if (jsoncand == NULL)
        dynarr_init(jsonkv_dynarr_t, *extrajson);
    else
        *extrajson = *jsoncand;
    return TRANS_OK;
}

typedef struct {
    char *url;
    charp_dynarr_t links;
    jsonkv_dynarr_t extrajson;
} linkentry_t;

dynarr_def(linkentry_t, linkentry_dynarr_t);

charp_dynarr_t getuniqlinks(linkentry_dynarr_t *links) {
    // Filters for unique links
    charp_hset_t uniqlinks_hset = hset_initi(charp_hset_t, hset_charp_hash, hset_charp_cmp);
    for (size_t i = 0; i < links->len; i++)
        for (size_t j = 0; j < links->data[i].links.len; j++)
            hset_add(uniqlinks_hset, links->data[i].links.data[j]);
    charp_dynarr_t uniqlinks = dynarr_initi(charp_dynarr_t);
    void *saveptr = NULL;
    char **link;
    while ((link = hset_iter(uniqlinks_hset, saveptr)) != NULL)
        dynarr_push(uniqlinks, *link);
    hset_destroy(uniqlinks_hset);
    return uniqlinks;
}

// Does some json stuff idek
jsonval_t links2json(charp_dynarr_t *uniqlinks, linkentry_dynarr_t *links, host_dynarr_t *hosts) {
    jsonkv_t kvpair;

    // Adds key/value pair for each host
    jsonkv_dynarr_t hostmap = dynarr_initi(jsonkv_dynarr_t);
    for (size_t i = 0; i < hosts->len; i++) {
        hostentry_t *curhost = &hosts->data[i];
        jsonkv_dynarr_t entrymap = dynarr_initi(jsonkv_dynarr_t);
        bool isup = curhost->visitcnt > curhost->totalfailurecnt;
        kvpair = (jsonkv_t){ .key = "up", .val = json_createbool(isup) };
        dynarr_push(entrymap, kvpair);
        kvpair = (jsonkv_t){ .key = "visitcnt", .val = json_createint(curhost->visitcnt) };
        dynarr_push(entrymap, kvpair);
        kvpair = (jsonkv_t){ .key = "failurecnt", .val = json_createint(curhost->totalfailurecnt) };
        dynarr_push(entrymap, kvpair);
        kvpair = (jsonkv_t){ .key = curhost->host, .val = json_createobj(entrymap) };
        dynarr_push(hostmap, kvpair);
    }

    // Adds key/value pair for each visited link
    jsonkv_dynarr_t linkmap = dynarr_initi(jsonkv_dynarr_t);
    for (size_t i = 0; i < links->len; i++) {
        linkentry_t *curentry = &links->data[i];
        jsonkv_dynarr_t entrymap = dynarr_initi(jsonkv_dynarr_t);
        jsonval_dynarr_t urlindicies = dynarr_initi(jsonval_dynarr_t);
        for (size_t j = 0; j < curentry->links.len; j++) {
            char *cururl = curentry->links.data[j];
            size_t urlind;
            for (urlind = 0; urlind < uniqlinks->len && strcmp(uniqlinks->data[urlind], cururl) != 0; urlind++)
                ;
            if (urlind >= uniqlinks->len)
                // Just a baby drinking coffee
                die("You should never see this message unless i fucked the hash set implementation");
            jsonval_t urlentry = json_createint(urlind);
            dynarr_push(urlindicies, urlentry);
        }
        kvpair = (jsonkv_t){ .key = "link_indicies", .val = json_createarr(urlindicies) };
        dynarr_push(entrymap, kvpair);
        kvpair = (jsonkv_t){ .key = "nlinks", .val = json_createint(urlindicies.len) };
        dynarr_push(entrymap, kvpair);
        for (size_t i = 0; i < curentry->extrajson.len; i++)
            dynarr_push(entrymap, curentry->extrajson.data[i]);
        kvpair = (jsonkv_t){ .key = curentry->url, .val = json_createobj(entrymap) };
        dynarr_push(linkmap, kvpair);
    }

    // Assembles object containing uniqlinks, hostmap, and linkmap
    jsonkv_dynarr_t parentmap = dynarr_initi(jsonkv_dynarr_t);
    jsonval_dynarr_t uniqlinks_json = dynarr_initi(jsonval_dynarr_t);
    for (size_t i = 0; i < uniqlinks->len; i++)
        dynarr_push(uniqlinks_json, json_createstr(uniqlinks->data[i]));
    kvpair = (jsonkv_t) { .key = "hosts", .val = json_createobj(hostmap) };
    dynarr_push(parentmap, kvpair);
    kvpair = (jsonkv_t) { .key = "urlindicies", .val = json_createarr(uniqlinks_json) };
    dynarr_push(parentmap, kvpair);
    kvpair = (jsonkv_t) { .key = "links", .val = json_createobj(linkmap) };
    dynarr_push(parentmap, kvpair);
    return json_createobj(parentmap);
}

void crawler(const char *seed, const crawlerconfig_t *config) {
    CURLU *curl_url_h = curl_url();
    if (curl_url_h == NULL) {
        fatal("CURL URL failed to initialize");
        return;
    }

    if ((signal(SIGINT, fatalsighandler) == SIG_ERR)) {
        //(signal(SIGSEGV, fatalsighandler) == SIG_ERR) ||
        //(signal(SIGTERM, fatalsighandler) == SIG_ERR)) {
        fatal("signal() failed: %s (code %d)", strerror(errno), errno);
        return;
    }

    crawlerstate_t state = { 0 };
    state.epollfd = -1; // For cleanup if error
    deque_init(charp_deque_t, state.to_crawl);
    dynarr_init(charp_dynarr_t, state.visited);
    dynarr_init(host_dynarr_t, state.hostentries);
    dynarr_init(int_dynarr_t, state.sockets);
    state.config = config;

    char *seed_buf = xmalloc(strlen(seed) + 1);
    strcpy(seed_buf, seed);
    deque_push_back(state.to_crawl, seed_buf);

    struct epoll_event *event_buf = xmalloc(sizeof(struct epoll_event) * EPOLL_BUF_SIZE);
    state.curl_multi_h = curl_multi_init();
    if (state.curl_multi_h == NULL) {
        fatal("CURL Multi failed to initialize"); 
        goto cleanup;
    }
    CURLMcode mc;
    if ((mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_SOCKETFUNCTION, socketcb)) ||
        (mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_SOCKETDATA, &state)) ||
        (mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_TIMERFUNCTION, timercb)) ||
        (mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_TIMERDATA, &state.timeout_ms)) ||
        (mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_MAX_HOST_CONNECTIONS, MAX_CONNECTIONS_PER_HOST)) ||
        (mc = curl_multi_setopt(state.curl_multi_h, CURLMOPT_MAX_TOTAL_CONNECTIONS, MAX_CONNECTIONS))) {
        fatal("curl_multi_setopt failed: %s (code %d)", curl_multi_strerror(mc), mc);
        goto cleanup;
    }

    state.epollfd = epoll_create1(0);
    if (state.epollfd < 0) {
        fatal("epoll_create1 failed: %s (code %d)", strerror(errno), errno);
        goto cleanup;
    }

    linkentry_dynarr_t linkentries = dynarr_initi(linkentry_dynarr_t);

    state.modules = config->enabledmodules;
    for (size_t i = 0; i < state.modules.len; i++) {
        int (*init)(crawlermodule_t *) = state.modules.data[i]->module.init;
        memset(&state.modules.data[i]->module, 0, sizeof(crawlermodule_t));
        state.modules.data[i]->module.init = init;
        state.modules.data[i]->module.init(&state.modules.data[i]->module);
    }

    size_t ntransfers = 0;
    int running_handles = 0;
    for(;;) {
        if (shoulddie > 0) {
            fatal("shoulddie = %d", shoulddie);
            break;
        }

        transfer_t *transfer;
        for (size_t i = 0; i < requestedreqs.len && ntransfers < MAX_CONNECTIONS; i++) {
            requestedreq_t *request = &requestedreqs.data[i];
            if (request->handle != NULL)
                continue;
            transfercode_t transcode = starttransfer(curl_url_h, &state, 
                    (char*)request->url, &transfer, true);
            if (transcode == TRANS_FATAL)
                break;
            else if (transcode != TRANS_OK)
                continue;
            transfer->cbdata.writecb_data.wasrequested = true;
            request->handle = transfer->handle;
            ntransfers++;
        }

        charp_dynarr_t stalled = dynarr_initi(charp_dynarr_t);
        for (size_t i = 0; i < state.to_crawl.len && ntransfers < MAX_CONNECTIONS; i++) {
            char *crawling = deque_pop_front(state.to_crawl);
            bool passed = true;
            for (size_t i = 0; i < state.modules.len; i++) {
                moduleentry_t *entry = state.modules.data[i];
                if (entry->module.filter == NULL)
                    continue;
                filterres_t res = entry->module.filter(entry->module.userdata, crawling);
                if (res != FILTER_PASS) {
                    passed = false;
                    char sanitized[100];
                    sanitize2ascii(sanitized, crawling, sizeof(sanitized));
                    if (res == FILTER_STALL) {
                        //debug("URL \"%s\" was stalled by module %s", sanitized, entry->name);
                        dynarr_push(stalled, crawling);
                        break;
                    }
                    else if (res == FILTER_REJECT) {
                        debug("URL \"%s\" was rejected by module %s", sanitized, entry->name);
                        // Don't add it to the visited list, in case the filter changes its mind
                        break;
                    }
                }
            }
            if (!passed)
                continue;
            transfercode_t transcode = starttransfer(curl_url_h, &state, crawling, &transfer, false);
            if (transcode == TRANS_FATAL)
                break;
            else if (transcode != TRANS_OK)
                continue;
            ntransfers++;
        }
        while (stalled.len > 0)
            deque_push_back(state.to_crawl, dynarr_pop(stalled));

        mc = perf_curl_multi_socket_action(state.curl_multi_h, -1, CURL_SOCKET_TIMEOUT, &running_handles);
        if (mc) {
            fatal("curl_multi_socket_action failed: %s (code %d)", curl_multi_strerror(mc), mc);
            goto cleanup;
        }

        if (running_handles == 0)
            continue;
        
        // Main add/remove transfer loop
        int availfds = perf_epoll_wait(state.epollfd, event_buf, EPOLL_BUF_SIZE, state.timeout_ms);
        if (availfds < 0) {
            fatal("epoll_wait failed: %s (code %d)", strerror(errno), errno);
            break;
        }
        transfer_t *trans = state.headtransfer;
        if (trans == NULL && state.to_crawl.len == 0) {
            info("No more URLs left to crawl");
            break;
        }

        // Tell CURL about any action connections
        if (availfds > 0) {
            for (int i = 0; i < availfds; i++) {
                struct epoll_event *connevent = &event_buf[i];
                int ev_bitmask = 0;
                // TODO: Check for errors on the descriptor
                if (connevent->events & EPOLLIN)
                    ev_bitmask |= CURL_CSELECT_IN;
                if (connevent->events & EPOLLOUT)
                    ev_bitmask |= CURL_CSELECT_OUT;
                CURLMcode mc;
                mc = perf_curl_multi_socket_action(state.curl_multi_h, connevent->data.fd, 
                        ev_bitmask, &running_handles);
                if (mc) {
                    fatal("curl_multi_socket_action failed: %s (code %d)", curl_multi_strerror(mc), mc);
                    goto cleanup;
                }
            }
        }
        else {
            CURLMcode mc;
            mc = perf_curl_multi_socket_action(state.curl_multi_h, CURL_SOCKET_TIMEOUT, -1, 
                    &running_handles);
            if (mc) {
                fatal("curl_multi_socket_action failed: %s (code %d)", curl_multi_strerror(mc), mc);
                break;
            }
        }
        // Process and prunes any finished connections
        CURLMsg *msg;
        int nmsgs;
        while ((msg = curl_multi_info_read(state.curl_multi_h, &nmsgs)) != NULL){
            // No other message types are currently defined, but in case they are
            if (msg->msg != CURLMSG_DONE) 
                continue;
            transfer_t *trans, *prev;
            for (trans = state.headtransfer, prev = NULL; 
                    trans != NULL; 
                    prev = trans, trans = trans->next
                )
                if (trans->handle == msg->easy_handle)
                    break;
            if (trans == NULL) {
                fatal("message handle not found, handle=%p, result=%d", 
                        (void*)msg->easy_handle, msg->data.result);
                goto cleanup;
            }

            // Check if the request was requested
            size_t requestind;
            for (requestind = 0; 
                requestind < requestedreqs.len &&
                    requestedreqs.data[requestind].handle != trans->handle;
                requestind++)
                ;
            reqcb_t cb = NULL;
            void *userdata = NULL;
            if (requestind < requestedreqs.len) {
                cb = requestedreqs.data[requestind].cb;
                userdata = requestedreqs.data[requestind].userdata;
                dynarr_remove(requestedreqs, requestind);
            }
            
            // Handle it and log that we visited it
            charp_dynarr_t links;
            jsonkv_dynarr_t extrajson;
            transfercode_t transcode = transferfinished(curl_url_h, &state, msg, trans, &links, 
                    &extrajson, cb, userdata);
            if (cb == NULL && transcode == TRANS_OK) {
                linkentry_t entry = { .url = trans->url, .links = links, .extrajson = extrajson };
                dynarr_push(linkentries, entry);
            }

            // Remove it
            if (prev == NULL)
                // Transfer was the head 
                state.headtransfer = trans->next;
            else
                // Transfer was not the head
                prev->next = trans->next;
            prev = trans;
            destroytransfer(state.curl_multi_h, trans);
            if (transcode == TRANS_FATAL)
                goto cleanup;
            ntransfers--;
        }
    }

    jsonval_t json;
    charp_dynarr_t uniqlinks;
cleanup:
    // Create and write out all of the json
    uniqlinks = getuniqlinks(&linkentries);
    json = links2json(&uniqlinks, &linkentries, &state.hostentries);
    json_write(stdout, &json);
    // Destroy all modules first bc they might have some save data
    for (size_t i = 0; i < state.modules.len; i++)
        if (state.modules.data[i]->module.destroy != NULL)
            state.modules.data[i]->module.destroy(state.modules.data[i]->module.userdata);
    // Destroy all links
    json_destroy(&json);
    for (size_t i = 0; i < linkentries.len; i++)
        dynarr_destroy(linkentries.data[i].links);
    dynarr_destroy(linkentries);
    for (size_t i = 0; i < uniqlinks.len; i++)
        xfree(uniqlinks.data[i]);
    dynarr_destroy(uniqlinks);
    // Destroy everything else
    for (size_t i = 0; i < state.hostentries.len; i++) {
        curl_slist_free_all(state.hostentries.data[i].headers);
        curl_free((void*)state.hostentries.data[i].host);
    }
    for (transfer_t *trans = state.headtransfer, *next; trans != NULL; trans = next) {
        next = trans->next;
        destroytransfer(state.curl_multi_h, trans);
    }
    if (state.curl_multi_h != NULL)
        curl_multi_cleanup(state.curl_multi_h);
    if (state.epollfd >= 0)
        close(state.epollfd);
    xfree(event_buf);
    curl_url_cleanup(curl_url_h);
    dynarr_destroy(state.visited);
    dynarr_destroy(state.hostentries);
    dynarr_destroy(state.sockets);
    dynarr_destroy(requestedreqs);
    deque_destroy(state.to_crawl);
}
