#include <curl/curl.h>
#include <tidybuffio.h>
#include <tidy.h>

#include "json.h"
#include "module.h"
#include "util.h"

#define MAX_HEADER_SIZE 256
#define MAX_TITLE_SIZE 256

bool getescapedheader(CURL *handle, const char *header, char **escaped) {
    struct curl_header *data;
    *escaped = NULL;
    CURLHcode res = curl_easy_header(handle, header, 0, CURLH_HEADER, -1, &data);
    if (res != CURLHE_OK) {
        if (res == CURLHE_BADINDEX || res == CURLHE_NOREQUEST || 
            res == CURLHE_NOREQUEST || res == CURLHE_MISSING)
            return true;
        error("curl_easy_handle() failed with code %d", res);
        return false;
    }
    int len = strlen(data->value);
    if (len > MAX_HEADER_SIZE) {
        error("max header size of %d bytes exceeded. header size is %d bytes", MAX_HEADER_SIZE, len);
        return false;
    }
    *escaped = sanitize2ascii_dyn(data->value, MAX_HEADER_SIZE * 4);
    if (escaped == NULL) {
        error("sanitize2ascii_dyn() failed");
        return false;
    }
    return true;
}

char *gettitle(TidyDoc doc, TidyNode node) {
    for (TidyNode child = tidyGetChild(node); child; child = tidyGetNext(child)) {
        ctmbstr name = tidyNodeGetName(child);
        if (!name)
            return NULL;
        if (strcmp(name, "title") == 0) {
            TidyNode textchild = tidyGetChild(child); // If conforming, should be text
            TidyBuffer buf;
            tidyBufInit(&buf);
            if (!tidyNodeGetText(doc, textchild, &buf)) {
                tidyBufFree(&buf);
                continue;
            }
            size_t len = strlen((char*)buf.bp);
            // tidy places a newline at the end of a title, so we have to be careful to get rid 
            // of it
            if (len <= 1) {
                tidyBufFree(&buf);
                continue;
            }
            char *ret = xmalloc(len+1);
            strcpy(ret, (char*)buf.bp);
            ret[len-1] = '\0';
            tidyBufFree(&buf);
            return ret;
        }
        char *ret = gettitle(doc, child);
        if (ret != NULL)
            return ret;
    }
    return NULL;
}

void dumpnode(TidyDoc doc, charp_dynarr_t *freearr, jsonkv_dynarr_t *json) {
    TidyNode head = tidyGetHead(doc);
    if (head) {
        char *title = gettitle(doc, head);
        if (title != NULL) {
            int len = strlen(title);
            if (len > MAX_TITLE_SIZE) {
                error("max title size of %d bytes exceeded. title size is %d bytes", MAX_TITLE_SIZE, len);
            }
            else {
                char *escaped = sanitize2ascii_dyn(title, MAX_TITLE_SIZE * 4);
                if (escaped == NULL) {
                    error("sanitize2ascii_dyn() failed");
                    xfree(escaped);
                }
                jsonkv_t kv = { .key = "title", .val = json_createstr(escaped) };
                dynarr_push(*json, kv);
                dynarr_push(*freearr, escaped);
            }
            xfree(title);
        }
    }
}

int mod_pagedata_onpagecomplete(void *userdata, pagecompletedata_t *data) {
    jsonkv_dynarr_t *jsonarr = 
        searchextradata(EXTRA_JSON, "json", data->extradata.data, data->extradata.len);
    if (jsonarr == NULL) {
        jsonkv_dynarr_t *arr = xmalloc(sizeof(jsonkv_dynarr_t));
        dynarr_init(jsonkv_dynarr_t, *arr);
        extradata_t entry = {
            .type = EXTRA_JSON,
            .key = "json",
            .val = arr,
        };
        dynarr_push(data->extradata, entry);
        jsonarr = (jsonkv_dynarr_t*)dynarr_get(data->extradata, data->extradata.len-1)->val;
    }

    charp_dynarr_t *freearr = userdata;
    char *headernames[] = {"Content-Type", "Last-Modified", "ETag"};
    for (size_t i = 0; i < array_size(headernames); i++) {
        char *escaped;
        if (!getescapedheader(data->handle, headernames[i], &escaped)) {
            error("appendheader() failed with header %s", headernames[i]);
            continue;
        }
        if (escaped == NULL)
            continue;
        dynarr_push(*freearr, escaped);
        jsonkv_t kv = { .key = headernames[i], .val = json_createstr(escaped) };
        dynarr_push(*jsonarr, kv);
    }

    TidyDoc doc = searchextradata(EXTRA_TIDY, "tidyDoc", data->extradata.data, data->extradata.len);
    if (doc == NULL)
        debug("url %s contained no tidy data, skipping document parsing...", data->url);
    else
        dumpnode(doc, freearr, jsonarr);
    return 0;
}

int mod_pagedata_destroy(void *userdata) {
    charp_dynarr_t *freearr = userdata;
    for (size_t i = 0; i < freearr->len; i++)
        xfree(freearr->data[i]);
    dynarr_destroy(*freearr);
    xfree(freearr);
    return 0;
}

int mod_pagedata_init(crawlermodule_t *entry) {
    *entry = (crawlermodule_t) {
        .userdata = xmalloc(sizeof(charp_dynarr_t)),
        .init = entry->init,
        .destroy = mod_pagedata_destroy,
        .onpagecomplete = mod_pagedata_onpagecomplete,
    };
    *(charp_dynarr_t*)entry->userdata = dynarr_initi(charp_dynarr_t);
    return 0;
}
