#include <curl/curl.h>
#include <ctype.h>
#include <tidybuffio.h>
#include <tidy.h>

#include "module.h"
#include "util.h"

#define MAX_LINK_LEN 512

bool islinksafe(char c) {
    return ('a' <= c && c <= 'z') ||
        ('A' <= c && c <= 'Z') ||
        ('0' <= c && c <= '9') ||
        c == '&' || c == '$' || c == ',' || c == '/' || 
        c == ':' || c == ';' || c == '=' || c == '?' || 
        c == '@' || c == '#' || c == '%' || c == '~' ||
        c == '_' || c == '-' || c == '(' || c == ')' ||
        c == '.';
}

char *relative2absolute(CURLU *curl_url_h, const char *parent, const char *relative) {
    CURLUcode url_res;
    char *curl_abs_link = NULL;
    url_res = curl_url_set(curl_url_h, CURLUPART_URL, parent, 0);
    if (url_res != CURLUE_OK) {
        char sanitized[100];
        sanitize2ascii(sanitized, relative, sizeof(sanitized));
        error("Parent URL parsing failed for \"%s\": %s", 
                sanitized, curl_url_strerror(url_res));
        return NULL;
    }
    url_res = curl_url_set(curl_url_h, CURLUPART_URL, relative, 0);
    if (url_res != CURLUE_OK){
        char sanitized[100];
        sanitize2ascii(sanitized, relative, sizeof(sanitized));
        error("Relative URL parsing failed for \"%s\": %s", 
                sanitized, curl_url_strerror(url_res));
        return NULL;
    } 
    url_res = curl_url_get(curl_url_h, CURLUPART_URL, &curl_abs_link, CURLU_PUNYCODE);
    if (url_res != CURLUE_OK) {
        error("Full URL parsing failed: %s", curl_url_strerror(url_res));
        return NULL;
    }
    char *ret = xmalloc(strlen(curl_abs_link) + 1);
    strcpy(ret, curl_abs_link);
    curl_free(curl_abs_link);
    return ret;
}

size_t geturlcutlen(const char *url, size_t len) {
    char *hash = memchr(url, '#', len);
    size_t hlen = hash == NULL ? len : hash - url;
    char *ques = memchr(url, '?', len);
    size_t qlen = ques == NULL ? len : ques - url;
    return min(hlen, qlen);
}

char *parselink(CURLU *curl_url_h, const char *parent, const char *child, size_t nchild) {
    size_t cutlen = geturlcutlen(child, nchild);
    if (cutlen < 1)
        return NULL;
    char *linkbuf = xmalloc(cutlen + 1);
    memcpy(linkbuf, child, cutlen);
    linkbuf[cutlen] = '\0';
    char *abslink = relative2absolute(curl_url_h, parent, linkbuf);
    if (abslink != NULL) {
        char sanitized[100];
        sanitize2ascii(sanitized, abslink, sizeof(sanitized));
        debug("found link: %s", sanitized);
    }
    xfree(linkbuf);
    return abslink;
}

int parsehrefs(CURLU *curl_url_h, const char *url, const char *page, size_t npage, charp_dynarr_t *ret) {
    const char *href = "href";
    size_t href_len = strlen(href), j = 0, linklen = 0;
    int state = 0, linkcnt = 0;
    // Probably a really good job for regex but wtvr
    for (size_t i = 0; i < npage; i++) {
        if (state == 0) {
            if (j == href_len) {
                state += 1;
                i -= 1;
            }
            if (tolower(page[i]) != href[j])
                j = 0;
        }
        else if (state == 1 && page[i] == '=') {
            state += 1;
        }
        else if (state == 2 && page[i] == '"') {
            state += 1;
            linklen = 0;
        }
        else if (state == 3) {
            if (page[i] == '"') {
                state = 0;
                if (linklen > 0) {
                    char *link = parselink(curl_url_h, url, page + i - linklen, linklen);
                    if (link != NULL) {
                        dynarr_push(*ret, link);
                        linkcnt++;
                    }
                }
            }
            if (!islinksafe(page[i]) || linklen >= MAX_LINK_LEN)
                state = 0;
            linklen += 1;
        }
        else if (page[i] != ' ' && page[i] != '\t' && page[i] != '\n') {
            state = 0;
        }
        j++;
    }
    return linkcnt;
}

char *tolower_s(const char *in) {
    size_t len = strlen(in);
    char *ret = xmalloc(len+1);
    for (size_t i = 0; i < len; i++)
        ret[i] = tolower(in[i]);
    ret[len] = '\0';
    return ret;
}

int parseattrs(CURLU *curl_url_h, const char *url, TidyNode node, charp_dynarr_t *links) {
    ctmbstr name = tidyNodeGetName(node);
    const char *texttags[] = { "p", "t", "span", "a" };
    char *namelower = tolower_s(name);
    bool found = false;
    for (size_t i = 0; !found && i < array_size(texttags); i++)
        if (strcmp(namelower, texttags[i]) == 0)
            found = true;
    xfree(namelower);
    if (!found)
        return 0;
    // Parse attributes for links
    int linkcnt = 0;
    for (TidyAttr attr = tidyAttrFirst(node); attr; attr = tidyAttrNext(attr)) {
        ctmbstr attrname = tidyAttrName(attr);
        if (attrname == NULL) {
            char sanitized[100];
            sanitize2ascii(sanitized, url, sizeof(sanitized));
            error("empty attr name for url \"%s\"", sanitized);
            continue;
        }
        char *attrnamelower = tolower_s(attrname);
        ctmbstr val = tidyAttrValue(attr);
        if (strcmp(attrnamelower, "href") != 0 || val == NULL) {
            xfree(attrnamelower);
            continue;
        }
        xfree(attrnamelower);
        // extract the link
        char *link = parselink(curl_url_h, url, val, strlen(val));
        if (link == NULL)
            continue;
        dynarr_push(*links, link);
        linkcnt++;
    }
    return linkcnt;
}

int parsenode(CURLU *curl_url_h, const char *url, TidyNode node, charp_dynarr_t *links) {
    int linkcnt = 0;
    for (TidyNode child = tidyGetChild(node); child; child = tidyGetNext(child)) {
        ctmbstr name = tidyNodeGetName(child);
        if (name == NULL) {
            // Node is probably text, for now do nothing
            // TODO: Parse plain text links
        }
        else {
            linkcnt += parseattrs(curl_url_h, url, child, links);
        }
        linkcnt += parsenode(curl_url_h, url, child, links);
    }
    return linkcnt;
}

int mod_parse_onpagecomplete(void *userdata, pagecompletedata_t *data) {
    CURLU *curl_url_h = userdata;
    TidyDoc doc = searchextradata(EXTRA_TIDY, "tidyDoc", data->extradata.data, data->extradata.len);
    if (doc == NULL) {
        debug("No tidied document found. Parsing hrefs...");
        parsehrefs(curl_url_h, data->url, data->page, data->npage, data->parsedlinks);
    } 
    else {
        TidyNode body = tidyGetBody(doc);
        if (!body)
            return 0;
        parsenode(curl_url_h, data->url, body, data->parsedlinks);
    }
    return 0;
}

int mod_parse_destroy(void *userdata) {
    CURLU *curl_url_h = userdata;
    curl_url_cleanup(curl_url_h);
    return 0;
}

int mod_parse_init(crawlermodule_t *entry) {
    *entry = (crawlermodule_t) {
        .init = entry->init,
        .onpagecomplete = mod_parse_onpagecomplete,
        .destroy = mod_parse_destroy,
        .userdata = curl_url(),
    };
    return 0;
}
