#include <string.h>
#include <stdlib.h>
#include <curl/curl.h>
#include <time.h>
#include <errno.h>

#include "module.h"
#include "util.h"

typedef struct hostdata {
    char *host;
    time_t crawldelay_ms;
    charp_dynarr_t prefixes;
    struct hostdata *next;
    struct timespec lastcrawled;
} hostdata_t;

typedef struct {
    hostdata_t *rules;
    CURLU *curl_url_h;
} state_t;

bool isprefixed(const char *str, const char *prefix) {
    size_t prefixlen = strlen(prefix);
    return prefixlen > strlen(str) ? false : memcmp(str, prefix, prefixlen) == 0;
}

void parse_robots_txt(hostdata_t *rules, char *page) {
    for (char *ptr = page; *ptr != '\0'; ptr++)
        if (*ptr == '\t')
            *ptr = ' ';
    char *linesave, *line;
    bool forus = true;
    // page is guarenteed to be terminated by a null byte
    while ((line = strtok_r(page, "\n", &linesave)) != NULL) {
        page = NULL;
        char *comment;
        if ((comment = memchr(line, '#', strlen(line))) != NULL)
            *comment = '\0';
        char *wssave, *func;
        if ((func = strtok_r(line, " ", &wssave)) == NULL)
            continue;
        char *tmp = xmalloc(strlen(func)+1);
        strcpy(tmp, func);
        func = tmp;
        char *arg;
        if ((arg = strtok_r(NULL, " ", &wssave)) == NULL)
            continue;

        if (strcmp(func, "User-agent:") == 0) {
            forus = strcmp(arg, "*") == 0 || isprefixed(arg, "AvaBot");
        }
        else if (strcmp(func, "Disallow:") == 0) {
            if (!forus) {
                xfree(func);
                continue;
            }
            if (strcmp(arg, "*") == 0) {
                dynarr_push(rules->prefixes, "/");
            }
            else {
                char *buf = xmalloc(strlen(arg)+1);
                strcpy(buf, arg);
                dynarr_push(rules->prefixes, buf);
            }
        }
        else if (strcmp(func, "Crawl-delay:") == 0) {
            if (!forus) {
                xfree(func);
                continue;
            }
            rules->crawldelay_ms = (unsigned long)(1000.0*atof(arg))/1000;
        }
        xfree(func);
    }
}

void robots_txt_cb(void *cbdata, const char *url, char *page, size_t npage, CURL *handle) {
    state_t *state = cbdata;
    hostdata_t *rules = state->rules;
    dynarr_init(charp_dynarr_t, rules->prefixes);
    rules->crawldelay_ms = 100;
    char *curl_path;
    CURLUcode rc;
    if ((rc = curl_url_set(state->curl_url_h, CURLUPART_URL, url, 0)) != CURLUE_OK ||
        (rc = curl_url_get(state->curl_url_h, CURLUPART_PATH, &curl_path, 0)) != CURLUE_OK) {
        error("curl_url operation failed: %s (code %d)", curl_url_strerror(rc), rc);
        return;
    }
    char *ctype;
    CURLcode ec;
    if ((ec = curl_easy_getinfo(handle, CURLINFO_CONTENT_TYPE, &ctype)) != CURLE_OK) {
        error("curl_easy_getinfo failed: %s (code %d)", curl_easy_strerror(ec), ec);
        return;
    }
    bool ctypegood = ctype == NULL ? false : isprefixed(ctype, "text/plain");
    if (!ctypegood || page == NULL) {
        if (strcmp(curl_path, "robots.txt") == 0) {
            char *curl_newurl;
            const char *newpath = ".well-known/robots.txt";
            if ((rc = curl_url_set(state->curl_url_h, CURLUPART_PATH, newpath, 0)) != CURLUE_OK ||
                (rc = curl_url_get(state->curl_url_h, CURLUPART_URL, &curl_newurl, 0)) != CURLUE_OK) {
                error("curl_url operation failed: %s (code %d)", curl_url_strerror(rc), rc);
                curl_free(curl_path);
                return;
            }
            char *newurl = xmalloc(strlen(curl_newurl)+1);
            strcpy(newurl, curl_newurl);
            curl_free(curl_newurl);
            makerequest(newurl, robots_txt_cb, cbdata);
        }
        return;
    }
    parse_robots_txt(rules, page);
    xfree(state); // state is cb owned
}

filterres_t mod_robots_filter(void *userdata, const char *url) {
    state_t *state = userdata;
    char *curl_host, *curl_path;
    CURLUcode rc;
    if ((rc = curl_url_set(state->curl_url_h, CURLUPART_URL, url, 0) != CURLUE_OK) ||
        (rc = curl_url_get(state->curl_url_h, CURLUPART_HOST, &curl_host, 0) != CURLUE_OK) ||
        (rc = curl_url_get(state->curl_url_h, CURLUPART_PATH, &curl_path, 0) != CURLUE_OK)) {
        char sanitized[100];
        sanitize2ascii(sanitized, url, sizeof(sanitized));
        error("curl_url operation failed for \"%s\": %s (code %d)", sanitized, 
                curl_url_strerror(rc), rc);
        return FILTER_PASS;
    }
    if (curl_host == NULL) {
        error("curl_host == NULL");
        return FILTER_PASS;
    }
    hostdata_t *rules = state->rules, *prev = NULL;
    for (; rules != NULL &&
            strcmp(rules->host, curl_host) != 0;
            prev = rules, rules = rules->next)
        ;
    filterres_t ret;
    if (rules == NULL) {
        char *curl_url;
        if ((rc = curl_url_set(state->curl_url_h, CURLUPART_PATH, "robots.txt", 0) != CURLUE_OK) ||
            (rc = curl_url_get(state->curl_url_h, CURLUPART_URL, &curl_url, 0) != CURLUE_OK)) {
            char sanitized[100];
            sanitize2ascii(sanitized, url, sizeof(sanitized));
            error("curl_url operation failed for \"%s\": %s (code %d)", sanitized, 
                    curl_url_strerror(rc), rc);
            ret = FILTER_PASS;
            goto cleanup;
        }
        if (curl_url == NULL) {
            error("curl_url == NULL");
            ret = FILTER_PASS;
            goto cleanup;
        }
        char *host = xmalloc(strlen(curl_host)+1);
        strcpy(host, curl_host);
        hostdata_t *newrules = xmalloc(sizeof(hostdata_t));
        *newrules = (hostdata_t) { .host = host, 0 };
        if (prev == NULL)
            state->rules = newrules;
        else
            prev->next = newrules;
        char *url = xmalloc(strlen(curl_url)+1);
        strcpy(url, curl_url);
        state_t *cbdata = xmalloc(sizeof(state_t));
        cbdata->rules = newrules;
        cbdata->curl_url_h = state->curl_url_h;
        debug("cbdata = %p", cbdata);
        makerequest(url, robots_txt_cb, cbdata);
        curl_free(curl_url);
        ret = FILTER_STALL;
    }
    else {
        if (rules->prefixes.data == NULL) { 
            // robots.txt request hasn't finished
            ret = FILTER_STALL;
            goto cleanup;
        }
        struct timespec now;
        if (clock_gettime(CLOCK_MONOTONIC, &now) < 0) {
            char *err = strerror(errno);
            error("clock_gettime failed: %s (code %d)", err, errno);
            ret = FILTER_STALL;
            goto cleanup;
        }
        time_t diff_ms = now.tv_sec * 1000 + now.tv_nsec / 1000000 - 
            rules->lastcrawled.tv_sec * 1000 - rules->lastcrawled.tv_nsec / 1000000;
        if (diff_ms < rules->crawldelay_ms) {
            ret = FILTER_STALL;
            goto cleanup;
        }
        rules->lastcrawled = now;
        ret = FILTER_PASS;
        for (size_t i = 0; i < rules->prefixes.len; i++) {
            if (isprefixed(curl_path, rules->prefixes.data[i])) {
                ret = FILTER_REJECT;
                break;
            }
        }
    }
cleanup:
    curl_free(curl_path);
    curl_free(curl_host);
    return ret;
}

int mod_robots_destroy(void *userdata) {
    state_t *state = userdata;
    curl_url_cleanup(state->curl_url_h);
    hostdata_t *cur = state->rules, *next;
    for (; cur != NULL; cur = next) {
        next = cur->next;
        dynarr_destroy(cur->prefixes);
        xfree(cur);
    }
    xfree(state);
    return 0;
}

int mod_robots_init(crawlermodule_t *entry) {
    state_t *state = xmalloc(sizeof(state_t));
    *state = (state_t) {
        .rules = NULL,
        .curl_url_h = curl_url(),
    };
    *entry = (crawlermodule_t) {
        .userdata = state,
        .init = entry->init,
        .filter = mod_robots_filter,
        .destroy = mod_robots_destroy,
    };
    return 0;
}
