#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdbool.h>
#include <pthread.h>
#include <unistd.h>
#include <stdint.h>

#include "util.h"

const char *last2path(const char *path) {
    const char *prev2 = NULL, *prev = NULL, *cur = path;
    for(;;) {
        prev2 = prev;
        prev = cur;
        cur = memchr(cur, '/', strlen(cur));
        if (cur++ == NULL)
            break;
    }
    if (prev2 == NULL)
        return path;
    return prev2;
}

void volog(loglevel_t level, const char *file, int line, const char *fmt, va_list ap) {
    char *cc;
    switch (level) {
    case LEVEL_DEBUG:
        cc = "\x1b[0mDEBUG";
        break;
    case LEVEL_INFO:
        cc = "\x1b[0mINFO";
        break;
    case LEVEL_WARN:
        cc = "\x1b[1;35mWARN";
        break;
    case LEVEL_ERROR:
        cc = "\x1b[1;31mERROR";
        break;
    case LEVEL_FATAL:
        cc = "\x1b[38;5;124mFATAL";
        break;
    }
    if (file == NULL)
        fprintf(stderr, "%s (unknown file): ", cc);
    else
        fprintf(stderr, "%s (%s:%d): ", cc, last2path(file), line);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\x1b[0m\n");
}

void olog(loglevel_t level, const char *file, int line, const char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    volog(level, file, line, fmt, ap);
    va_end(ap);
}

void die(const char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    volog(LEVEL_FATAL, NULL, 0, fmt, ap);
    va_end(ap);
    exit(1);
}

void *xmalloc(size_t size) {
    void *ret = malloc(size);
    if (ret == NULL)
        die("xmalloc failed: %s", strerror(errno));
    return ret;
}

void *xcalloc(size_t nmemb, size_t size) {
    void *ret = calloc(nmemb, size);
    if (ret == NULL)
        die("xcalloc failed: %s", strerror(errno));
    return ret;
}

void *xrealloc(void *ptr, size_t size) {
    void *old = ptr;
    ptr = realloc(ptr, size);
    if (ptr == NULL)
        die("xrealloc failed: %s", strerror(errno));
    return ptr;
}

void xfree(void *ptr) {
    free(ptr);
}

size_t parityhash(const void *data, size_t ndata) {
    size_t sum = 0;
    for (size_t i = 0; i < ndata; i++)
        sum += *((unsigned char*)data + i);
    return sum;
}

char *sanitize2ascii_dyn(const char *inp, size_t maxlen) {
    size_t nout = strlen(inp), out_ind = 0;
    if (nout > maxlen)
        return NULL;
    char *out = xmalloc(nout);
    char tmp[5];
    size_t inp_ind;
    for (inp_ind = 0; inp[inp_ind] != '\0'; inp_ind++) {
        if (inp[inp_ind] < 0x20 || inp[inp_ind] > 0x7e) {
            if (out_ind + 5 >= maxlen)
                break;
            if (out_ind + 5 >= nout) {
                nout *= 2;
                out = xrealloc(out, nout);
            }
            snprintf(tmp, sizeof(tmp), "\\x%02x", inp[inp_ind]);
            memcpy(out + out_ind, tmp, 4);
            out_ind += 4; // needs to accomidate "\xXX"
        }
        else if (inp[inp_ind] == '\\' || inp[inp_ind] == '"') {
            if (out_ind + 3 >= maxlen)
                break;
            if (out_ind + 3 >= nout) {
                nout *= 2;
                out = xrealloc(out, nout);
            }
            out[out_ind++] = '\\';
            out[out_ind++] = inp[inp_ind];
        }
        else {
            if (out_ind + 2 >= maxlen)
                break;
            if (out_ind + 2 >= nout) {
                nout *= 2;
                out = xrealloc(out, nout);
            }
            out[out_ind++] = inp[inp_ind];
        }
    }
    if (inp[inp_ind] != '\0') {
        xfree(out);
        return NULL;
    }
    out[out_ind] = '\0';
    return out;
}

size_t sanitize2ascii(char *out, const char *inp, size_t outsize) {
    size_t out_ind = 0;
    char tmp[5];
    for (size_t inp_ind = 0; inp[inp_ind] != '\0'; inp_ind++) {
        if (inp[inp_ind] < 0x20 || inp[inp_ind] > 0x7e) {
            if (out_ind + 5 >= outsize)
                break;
            snprintf(tmp, sizeof(tmp), "\\x%02x", inp[inp_ind]);
            memcpy(out + out_ind, tmp, 4);
            out_ind += 4; // needs to accomidate "\xXX"
        }
        else if (inp[inp_ind] == '\\' || inp[inp_ind] == '"') {
            if (out_ind + 3 >= outsize)
                break;
            out[out_ind++] = '\\';
            out[out_ind++] = inp[inp_ind];
        }
        else {
            if (out_ind + 2 >= outsize)
                break;
            out[out_ind++] = inp[inp_ind];
        }
    }
    out[out_ind] = '\0';
    return out_ind;
}

// This is the sample c implementation of MurmurHash taken from Wikipedia. Credit to the Wikipedia article 
// of MurmurHash and whoever made the sample implementation on the page. 
// https://en.wikipedia.org/w/index.php?title=MurmurHash&oldid=1218923262 accessed on 2024-06-02T18+00:00
// ----- BEGIN WIKIPEDIA SAMPLE CODE -----
static inline uint32_t murmur_32_scramble(uint32_t k) {
    k *= 0xcc9e2d51;
    k = (k << 15) | (k >> 17);
    k *= 0x1b873593;
    return k;
}
uint32_t murmur3_32(const uint8_t* key, size_t len, uint32_t seed)
{
	uint32_t h = seed;
    uint32_t k;
    /* Read in groups of 4. */
    for (size_t i = len >> 2; i; i--) {
        // Here is a source of differing results across endiannesses.
        // A swap here has no effects on hash properties though.
        memcpy(&k, key, sizeof(uint32_t));
        key += sizeof(uint32_t);
        h ^= murmur_32_scramble(k);
        h = (h << 13) | (h >> 19);
        h = h * 5 + 0xe6546b64;
    }
    /* Read the rest. */
    k = 0;
    for (size_t i = len & 3; i; i--) {
        k <<= 8;
        k |= key[i - 1];
    }
    // A swap is *not* necessary here because the preceding loop already
    // places the low bytes in the low places according to whatever endianness
    // we use. Swaps only apply when the memory is copied in a chunk.
    h ^= murmur_32_scramble(k);
    /* Finalize. */
	h ^= len;
	h ^= h >> 16;
	h *= 0x85ebca6b;
	h ^= h >> 13;
	h *= 0xc2b2ae35;
	h ^= h >> 16;
	return h;
}
// ----- END WIKIPEDIA SAMPLE CODE -----

bool hset_charp_cmp(char **lhs, char **rhs) {
    return strcmp(*lhs, *rhs) == 0;
}

size_t hset_charp_hash(char **str) {
    // u32 -> u64?
    return (size_t)murmur3_32(*str, strlen(*str), 0x9747b28c);
}
